### NOH-modular Comper firmware v1.0 - licensed under a CC-BY-NC-SA 4.0 license - https://creativecommons.org/licenses/by-nc-sa/4.0/
################################
### Libraries Import
################################
from machine import Pin, ADC
from random import getrandbits
from math import ceil
from micropython import alloc_emergency_exception_buf

alloc_emergency_exception_buf(100)

################################
### Pin Set-up and Functions
################################
density_CV = ADC(26)
time_signature = ADC(27)
brtm = Pin(1,Pin.IN)
brst = Pin(2,Pin.IN)
clock = Pin(4,Pin.IN)
lrtm = Pin(28,Pin.OUT)
lrst = Pin(3,Pin.OUT)
chord_out = Pin(29,Pin.OUT)
eight_out = Pin(7,Pin.OUT)
bass_out = Pin(6,Pin.OUT)
metre_out = Pin(0,Pin.OUT)

################################
### Stored Gate Loops
################################
bank = [
        #--------------------------------------------
        #3/4 rhythms - swing
        [[0xc0000007,0xc0000010c0001010,0xc0001010,0xc0010010c0000047,0xc0010407c0030001,0xc00101c0,0xc00001f1,
          0xc0001c07,0xc00007f0,0xc001dc00,0xc00071c0,0xc0010470,0xc0010445,0xc0001c10c001fc1f,0xc00071d0,
          0xc00011ff,0xc0007c1f,0xc0007770,0xc00007ffc003f1f0,0xc001ddc7],
        # - straight
        [0x80000007,0x800000c3,0x800030c0,0x800180c080003018,0x80003018,0x8000061b,0x80018618,
         0x800186038003f01f,0x8001b01f,0x800186db80038618],
        # - user
        []],
        
        #--------------------------------------------
        #4/4 rhythms - swing
        [[0xc0000007,0xc0400400,0xc0000407,0xc01c0000c0010451,0xc001f010c0000407,0xc0000471,0xc001f010,
          0xc01f01c0c0011410,0xc01f0011,0xc0000447c01c7410,0xc0010407c0777001,0xc00001ffc0c101c7,0xc00111ffc0c1f010,
          0xc01f0407c0fc7c1f,0xc00111ffc01c7fc7],
        # - straight
        [0x80000007,0x800030c380e180c0,0x80018618806030c3,0x800c3e00806187c0,0x800c30d88061801f,0x8061b0d8800030d8,
         0x800c36188061b0c3,0x800f8618806187c3,0x800c30f8806187c3,0x800f87c3,0x800ffe03,0x800f87c380e387c3,
         0x800df61f800df0df,0x80e1b61f80fc3e1b,0x800df0df80777777],
        # - user
        []],
        
        #--------------------------------------------
        #7/8 rhythms - straight
        [[0x80000007,0x80000003801c001f,0x800180c0,0x800000f8,0x800007ff800000c3,0x80018618,0x800c30c3,0x800c06c3,
          0x800c061f,0x800007ff,0x80003e1f,0x8001f01f,0x8001f0c3,0x800186c3801c361f,0x8001fe1f801ff0ff],
        # - straight
        [],
        # - user
        []]]

################################
### Global Variables
################################
chord_buffer = 0xFFFFFFFF
eight_buffer = 0xFFFFFFFF
bass_buffer = 0xFFFFFFFF
metre_buffer = 0x7
main_counter = [0,0]

time_sig = 0
previous_rtm=[[-1,0,0],False]
ppqn = 24

brtm_flag = False
brst_flag = False
brst_pushed = False
clock_flag = False

################################
### Pot Calibration
################################
pot_calibration=[[0,200],[1,2130],[2,9385],[3,19980],[4,32175],[5,47170],[6,52050]]

################################
### time_signature Functions
################################
def Readtime_sig():
    a = (time_signature.read_u16()>>14)
    return a - (a==3)

################################
### Control Voltage Functions
################################
def ReadPot():
    global pot_calibration
    
    # read and clip ADC value
    a = min(max(density_CV.read_u16(),pot_calibration[0][1]),pot_calibration[len(pot_calibration)-1][1]-1)
    
    # figure out which interpolation zone it lies in
    zone = 0
    for i in range(len(pot_calibration)-1):
        if pot_calibration[i][1] <= a <= pot_calibration[i+1][1]:
            zone = i
        else:
            pass
        
    # 1-D linear interpolation and return the ratio
    b = ((a-pot_calibration[zone][1])/(pot_calibration[zone+1][1]-pot_calibration[zone][1]))+pot_calibration[zone][0]
    return b/6
    
################################
### Sequences Functions
################################  
def reset_handler():
    global main_counter,brst_flag
    
    main_counter = [0,0]
    brst_flag = False
    lrst.value(0)
    
def update_buffers():
    global bank,chord_buffer,eight_buffer,time_sig,previous_rtm,brtm_flag,bass_buffer,ppqn
    
    time_sig = Readtime_sig()
    
    #figure out bank rhythm
    lens = [len(bank[time_sig][0]),len(bank[time_sig][1]),len(bank[time_sig][2])]
    pot_ratio = ReadPot()
    division=3-(lens[0]==0)-(lens[1]==0)-(lens[2]==0)
    section=int(pot_ratio*division)
    sub_section=int(((pot_ratio-(section*(1/division)))*division)*lens[section])
    
    #make it go to 2nd measure if needed
    if (bank[time_sig][section][sub_section]>>63) and (previous_rtm[0] == [time_sig,section,sub_section]):
        previous_rtm[1] = not previous_rtm[1]
        chord_buffer = bank[time_sig][section][sub_section]>>(32*previous_rtm[1])
    else:
        previous_rtm[1] = False
        chord_buffer = bank[time_sig][section][sub_section]
    previous_rtm[0]=[time_sig,section,sub_section]
    
    #8th + bass + random outputs
    #swung
    if ((chord_buffer&0xFFFFFFFF)>>30)&1:
        #random button
        if brtm_flag:
            rand = 0
            for i in range(12):
                rand = (rand<<2)|(getrandbits(1)*(((i-1)%3)!=0))
            chord_buffer = (chord_buffer>>24)<<24|rand
        else:
            pass
        #8th
        if ppqn == 4:
            eight_buffer = 0x80249249
        else:
            eight_buffer = 0xC1451451
        #bass
        bass_buffer = 0
        for i in range(2):
            bass_buffer = bass_buffer<<(12*i)|(getrandbits(4)>=15)<<10|(not(getrandbits(4)>=15))*7<<6|(getrandbits(4)>=15)<<4|0b111
            #quarter note always -> swing 1/16 -> quarter note 15/16 -> swing 1/16 -> [LOOP]
        
    #straight
    else:
        #random button
        if brtm_flag:
            rand = 0
            for i in range(8):
                rand = (rand<<3)|getrandbits(1)
            chord_buffer = (chord_buffer>>24)<<24|rand
        else:
            pass
        #8th
        eight_buffer = 0x80249249
        #bass
        if time_sig==0:#3/4
            bass_buffer = 0
            for i in range(2):
                bass_buffer = bass_buffer<<(9*i)|(getrandbits(4)>=15)<<6|0b111-((((i+1)%2)*getrandbits(1))*7)
                #quarter note 1/2 -> eigth note 1/16 -> quarter note always -> eigth note 1/16
                
        elif time_sig==1:#4/4
            bass_buffer = 0
            for i in range(2):
                bass_buffer = bass_buffer<<(12*i)|(getrandbits(4)>=1)<<9|(getrandbits(5)>=31)*15<<6|0b1111
                #quarter note always -> eigth note 1/32 -> eigth note 15/16 -> [LOOP]
                
        else:#7/8
            bass_buffer = 0
            for i in range(2):
                bass_buffer = bass_buffer<<(12*i)|(getrandbits(4)>=15)<<6|0b111
                #quarter note always -> quarter note 1/16 -> [LOOP]
        
def update_outputs(rst,btn):
    global chord_buffer,eight_buffer,bass_buffer,metre_buffer,main_counter,time_sig,brtm_flag,ppqn
    
    if rst:
        reset_handler()
    else:
        pass
    
    if (ppqn != 24) or ((ppqn == 24) and ((main_counter[0]%4)==0)):
        if main_counter[1]==0:
            main_counter=[0,0]
            update_buffers()
            lrtm.value(0)
            brtm_flag=False
        else:
            pass
        chord_out.value((chord_buffer>>main_counter[1])&1)
        eight_out.value((eight_buffer>>main_counter[1])&1)
        bass_out.value((bass_buffer>>main_counter[1])&1)
        metre_out.value((metre_buffer>>main_counter[1])&1)
    
    main_counter[0] += 1
    main_counter[1] = ceil(((main_counter[0])/(ppqn*4))*24)
    if main_counter[1] == (24-(3*(time_sig==2))-(6*(time_sig==0))):
        main_counter[1] = 0
    else:
        pass
    
################################
### Initial Clock Resolution Functions
################################
def save_ppqn(a):
    #write ppqn to the file
    save = open('ppqn.csv', 'w')
    save.write(str(a))
    save.close()
    
def load_ppqn():
    global ppqn
    
    #read and save ppqn value
    load = open('ppqn.csv', 'r')
    ppqn = int(load.read())
    load.close()
    
    #check if read ppqn is usable
    if ppqn==4 or ppqn==6:
        pass
    else:
        ppqn=24

def change_resolution():
    global ppqn
    
    #counter
    count = 0
    flip = False
    
    #light up to indicate the current mode
    lrtm.value(1)
    lrst.value(1)
    
    #wait for user  to make selection on time signature
    while (ReadPot()<0.7) or (not brtm.value()) or (not brst.value()):
        count = (count+1)%1000
        if count==0:
            flip = not flip
            lrtm.value(flip)
            lrst.value(flip)
    
    #get new clock resolution from time signature position
    change = Readtime_sig()
    if change==0:#3/4
        ppqn = 4
    elif change==2:#7/8
        ppqn = 6
    else: #4/4
        ppqn = 24
    save_ppqn(ppqn)

################################
### Initialisation of Clock Resolution
################################
load_ppqn()
    
if (Readtime_sig()==2) and (brtm.value()) and (brst.value()) and (ReadPot()<0.3):
    change_resolution()
else:
    pass

################################
### Clock IRQ
################################
def clock_handler(pin):
    global clock_flag
    
    if clock_flag:
        pass
    else:
        clock_flag=True
        update_outputs(brst_flag,brtm_flag)
        clock_flag=False
    
################################
### Initialisation
################################ 
time_sig = Readtime_sig()
lrtm.value(0)
lrst.value(0)
clock.irq(handler=clock_handler,trigger=Pin.IRQ_RISING)

################################
### Main Loop
################################

while True:
    if brtm.value() and not brtm_flag:
        lrtm.value(1)
        brtm_flag = True
    else:
        pass
    
    if brst.value():
        if not brst_pushed:
            brst_pushed = True
            brst_flag = True
            lrst.value(1)
        else:
            pass
    else:
        brst_pushed = False